#ifndef _OS_File_H
#define _OS_File_H

#include "Prefix.h"
#include <windows.h>

namespace EARTH {
namespace OS {
	class File {
	public:
		File()
		{
			mHandle = INVALID_HANDLE_VALUE;
		}

		~File()
		{
			Close();
		}

		status Open(const wchar_t *filename)
		{
			wchar_t directory[MAX_PATH];

			#if   defined(ENV_EARTHSOFT)
				::swprintf_s(directory, MAX_PATH, L"J:\\TS");
			#elif defined(ENV_FACTORY)
				::swprintf_s(directory, MAX_PATH, L"C:\\TS");
			#else
				::GetCurrentDirectory(MAX_PATH, directory);
			#endif

			wchar_t path[MAX_PATH];
			::swprintf_s(path, MAX_PATH, L"%s\\%s", directory, filename);

			_ASSERTE(mHandle == INVALID_HANDLE_VALUE);
			_ASSERTE(filename);
			mHandle = ::CreateFile(path, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_FLAG_NO_BUFFERING, NULL);
			if (mHandle == INVALID_HANDLE_VALUE) {
				return ::GetLastError();
			}

			return 0;
		}

		status Close()
		{
			if (mHandle != INVALID_HANDLE_VALUE) {
				BOOL b = ::CloseHandle(mHandle);
				if (b == FALSE) {
					return ::GetLastError();
				}
				mHandle = INVALID_HANDLE_VALUE;
			}

			return 0;
		}

		status Write(const void *ptr, uint size)
		{
			BOOL b;

			DWORD writeSize;
			b = ::WriteFile(mHandle, ptr, size, &writeSize, NULL);
			if (b == FALSE) {
				return ::GetLastError();
			}

			return 0;
		}

	protected:
		void *mHandle;
};

}
}

#endif
