#ifndef _OS_Thread_H
#define _OS_Thread_H

#include "Prefix.h"
#include <Windows.h>

namespace EARTH {
namespace OS {
	class Thread {
	public:
		typedef void (*Function)(void *);

		Thread(Function function, void *argument) : mFunction(function), mArgument(argument)
		{
			mThread = NULL;

			_ASSERTE(mThread == NULL);
			mThread = ::CreateThread(NULL, 0, ThreadFunction_, this, 0, NULL);
			if (mThread == NULL) { _ASSERTE(false); return; }
		}

		~Thread()
		{
			if (mThread) {
				DWORD result = ::WaitForSingleObject(mThread, INFINITE);
				_ASSERTE(result == WAIT_OBJECT_0);
				::CloseHandle(mThread);
				mThread = NULL;
			}
		}

		bool IsValid() const
		{
			return (mThread != NULL);
		}

		static void Sleep(ulong ns)
		{
			uint ms = static_cast<uint>(ns / (1000 * 1000));
			::Sleep(ms);
		}

	protected:
		const Function mFunction;
		void *const mArgument;
		HANDLE mThread;

		void ThreadFunction()
		{
			_ASSERTE(mFunction);
			mFunction(mArgument);
		}

		static uint __stdcall ThreadFunction_(void *argument)
		{
			Thread	*that = static_cast<Thread *>(argument);
			_ASSERTE(that);

			that->ThreadFunction();

			return 0;
		}
	};
}
}

#endif
