#ifndef _TS_Device_H
#define _TS_Device_H

#include "Prefix.h"
#include "EARTH_PT.h"
#include "TS_Buffer.h"
#include "TS_Utility.h"
#include "OS_Thread.h"

#if defined(ENV_EARTHSOFT) && defined(_DEBUG)
#include "EARTH_PT_.h"
#include "TS_Device2.h"
#endif

namespace EARTH {
namespace TS {
	class Device {
	public:
		Device()
		{
			mDevice = NULL;
			#if defined(ENV_EARTHSOFT) && defined(_DEBUG)
			mDevice2 = NULL;
			#endif
			mBuffer = NULL;

			mIsOpen = false;
			mIsValid = false;
		}

		~Device()
		{
			#if defined(ENV_EARTHSOFT) && defined(_DEBUG)
			delete_(mDevice2);
			#endif

			delete_(mBuffer);

			if (mDevice) {
				status status;

				if (mIsOpen) {
					status = mDevice->Close();
					if (Error(L"Device::Close()", status)) return;
					mIsOpen = false;
				}

				status = mDevice->Delete();
				if (Error(L"Device::Delete()", status)) return;

				mDevice = NULL;
			}
		}

		void Run(PT::Bus *bus, const PT::Bus::DeviceInfo *deviceInfo)
		{
			Init(bus, deviceInfo);
			if (mIsValid == false) return;

			Menu1();
		}

	protected:
		PT::Device *mDevice;

		#if defined(ENV_EARTHSOFT) && defined(_DEBUG)
		TS::Device2 *mDevice2;
		#endif

		Buffer *mBuffer;
		bool mIsOpen;
		bool mIsValid;

		void Init(PT::Bus *bus, const PT::Bus::DeviceInfo *deviceInfo)
		{
			status status;

			#if defined(ENV_EARTHSOFT) && defined(_DEBUG)
			PT::Device2 *device2;
			#endif

			status = bus->NewDevice(deviceInfo, &mDevice,
				#if defined(ENV_EARTHSOFT) && defined(_DEBUG)
				&device2
				#else
				NULL
				#endif
			);
			if (Error(L"Bus::NewDevice()", status)) return;

			#if defined(ENV_EARTHSOFT) && defined(_DEBUG)
			mDevice2 = new Device2(mDevice, device2, deviceInfo->PTn);
			#endif

			_ASSERTE(mDevice);
			status = mDevice->Open();
			if (Error(L"Device::Open()", status)) return;
			mIsOpen = true;

			status = mDevice->SetTunerPowerReset(PT::Device::TUNER_POWER_ON_RESET_ENABLE);
			if (Error(L"Device::SetTunerPowerReset(TUNER_POWER_ON_RESET_ENABLE)", status)) return;

			OS::Thread::Sleep((20)*1000*1000); // 20 ms ҂

			status = mDevice->SetTunerPowerReset(PT::Device::TUNER_POWER_ON_RESET_DISABLE);
			if (Error(L"Device::SetTunerPowerReset(TUNER_POWER_ON_RESET_DISABLE)", status)) return;

			OS::Thread::Sleep((1)*1000*1000); // 1 ms ҂

			for (uint i=0; i<2; i++) {
				status = mDevice->InitTuner(i);
				if (Error(L"Device::InitTuner()", status)) return;
			}

			for (uint i=0; i<2; i++) {
				for (uint j=0; j<2; j++) {
					status = mDevice->SetTunerSleep(i, static_cast<PT::Device::ISDB>(j), false);
					if (Error(L"Device::SetTunerSleep()", status)) return;
				}
			}

			for (uint i=0; i<2; i++) {
				for (uint j=0; j<2; j++) {
					status = mDevice->SetStreamEnable(i, static_cast<PT::Device::ISDB>(j), true);
					if (Error(L"Device::SetStreamEnable()", status)) return;
				}
			}

			// SetStreamGray() Őݒ肵l TS pPbg 189 oCgڂ̉ 3 rbgɔf܂B
			// ̃Tvł͐ϋɓIɎgĂ܂B
			for (uint i=0; i<2; i++) {
				for (uint j=0; j<2; j++) {
					status = mDevice->SetStreamGray(i, static_cast<PT::Device::ISDB>(j), 4 + i*2 + j);
					if (Error(L"Device::SetStreamGray()", status)) return;
				}
			}

			mIsValid = true;
		}

		void Menu1()
		{
			bool done = false;
			while (done == false) {
				bool transferEnable;
				status status = mDevice->GetTransferEnable(&transferEnable);
				if (Error(L"Device::GetTransferEnable()", status)) return;

				::wprintf(L"[foCXj[ (1^3y[W)]\n");
				::wprintf(L"0: LNB dݒ\n");
				::wprintf(L"1: `lXL\n");
				::wprintf(L"2: `lݒ\n");
				::wprintf(L"3: ISDB-S TS-ID ݒ\n");
				::wprintf(L"4: G[[gEG[JE^擾\n");
				::wprintf(L"5: Lv`%s\n", transferEnable ? L"~" : L"Jn");
				::wprintf(L"6:  2y[W\n");
				::wprintf(L"7:  3y[WځiHo׎ځj\n");
				::wprintf(L"8: foCX\n");
				::wprintf(L">");
				uint number = GetNumber(8);
				if (number == 8) {
					done = true;
				} else {
					switch (number) {
						case 0: SetLnbPower(); break;
						case 1: ScanChannel(); break;
						case 2: SetChannel(); break;
						case 3: SetTsId(); break;
						case 4: ErrorRateCounter(); break;
						case 5: Capture(!transferEnable); break;
						case 6: Menu2(); break;

						#if defined(ENV_EARTHSOFT) && defined(_DEBUG)
						case 7: mDevice2->Menu3(); break;
						#endif
					}
				}
			}
		}

		void SetLnbPower()
		{
			status status;

			PT::Device::LnbPower lnbPower;
			status = mDevice->GetLnbPower(&lnbPower);
			if (Error(L"Device::GetLnbPower()", status)) return;

			const wchar_t *LABEL[] = { L"It", L"+15V", L"+11V" };

			::wprintf(L"[LNB dݒ]\n");
			for (int i=0; i<sizeof(LABEL)/sizeof(*LABEL); i++) {
				::wprintf(L"%d: %s%s\n", i, LABEL[i], (i==lnbPower) ? L" (ݒl)" : L"");
			}
			::wprintf(L"3: LZ\n");
			::wprintf(L">");
			uint number = GetNumber(3);
			if (number == 3) return;

			status = mDevice->SetLnbPower(static_cast<PT::Device::LnbPower>(number));
			if (Error(L"Device::SetLnbPower()", status)) return;
		}

		void ScanChannel()
		{
			::wprintf(L"0: `[i[ԍ0\n");
			::wprintf(L"1: `[i[ԍ1\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int tuner = GetNumber(2);
			if (tuner == 2) return;

			::wprintf(L"0: ISDB-S\n");
			::wprintf(L"1: ISDB-T\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int isdb = GetNumber(2);
			if (isdb == 2) return;

			switch (isdb) {
				case 0: ScanS(tuner, 0,  23); break;
				case 1: ScanT(tuner, 0, 112); break;
			}
		}

		void SetChannel()
		{
			::wprintf(L"0: `[i[ԍ0\n");
			::wprintf(L"1: `[i[ԍ1\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int tuner = GetNumber(2);
			if (tuner == 2) return;

			::wprintf(L"0: ISDB-S\n");
			::wprintf(L"1: ISDB-T\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int isdb = GetNumber(2);
			if (isdb == 2) return;

			uint max = (isdb == PT::Device::ISDB_S) ? 35 : 112;

			::wprintf(L"`lԍ͂ĂB(͈:0`%d)\n", max);
			::wprintf(L">");

			uint channel;
			::wscanf_s(L"%d", &channel);
			if (max < channel) {
				channel = max;
			}

			status status;
			if (isdb == PT::Device::ISDB_S) {
				status = ScanS(tuner, channel, channel);
			} else {
				status = ScanT(tuner, channel, channel);
			}
		}

		void SetTsId()
		{
			::wprintf(L"0: `[i[ԍ0\n");
			::wprintf(L"1: `[i[ԍ1\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int tuner = GetNumber(2);
			if (tuner == 2) return;

			::wprintf(L"TS-ID ͂ĂB(͈:0x0000`0xffff)\n");
			::wprintf(L">0x");

			uint id;
			::wscanf_s(L"%x", &id);
			if (0xffff < id) {
				id = 0xffff;
			}

			status status = mDevice->SetIdS(tuner, id);
			if (Error(L"Device::SetIdS()", status)) return;
		}

		void ErrorRateCounter()
		{
			ErrorRate();
			ErrorCounter();
		}

		void ErrorRate()
		{
			::wprintf(L"([h\ŒꂽG[[g)\n");
			for (uint i=0; i<2; i++) {
				for (uint j=0; j<PT::Device::ISDB_COUNT; j++) {
					uint layerCount = (j == PT::Device::ISDB_S) ? PT::Device::LAYER_COUNT_S : PT::Device::LAYER_COUNT_T;
					for (uint k=0; k<layerCount; k++) {
						PT::Device::ErrorRate errorRate;
						mDevice->GetCorrectedErrorRate(
							i, static_cast<PT::Device::ISDB>(j), static_cast<PT::Device::LayerIndex>(k), &errorRate);

						uint numerator = errorRate.Numerator;
						uint denominator = errorRate.Denominator;
						uint e = 0;
						if (numerator) {
							while (numerator < denominator) {
								numerator *= 10;
								e++;
							}
						}
						double rate;
						if (denominator) {
							rate = static_cast<double>(numerator) / denominator;
						} else {
							rate = 0;
						}

						wchar_t isdb = (j == PT::Device::ISDB_S) ? L'S' : L'T';
						wchar_t layer = (j == PT::Device::ISDB_S) ? ((k == 0) ? L'' : L'') : static_cast<wchar_t>(L'A'+k);
						const wchar_t *space = (j == PT::Device::ISDB_S) ? L"" : L" ";
						::wprintf(L" `[iԍ:%d ISDB-%c %c%sKw %.2f~10^-%d (%d/%d)\n",
							i, isdb, layer, space, rate, e, errorRate.Numerator, denominator);
					}
				}
			}
		}

		void ErrorCounter()
		{
			::wprintf(L"([h\ŒȂ TS pPbg)\n");
			for (uint i=0; i<2; i++) {
				::wprintf(L" `[iԍ:%d", i);
				for (uint j=0; j<PT::Device::ISDB_COUNT; j++) {
					uint count;
					mDevice->GetErrorCount(i, static_cast<PT::Device::ISDB>(j), &count);

					wchar_t isdb = (j == PT::Device::ISDB_S) ? L'S' : L'T';
					::wprintf(L" ISDB-%c:0x%06x", isdb, BIT_SHIFT_MASK(count, 0, 24));
				}
				::wprintf(L"\n");
			}
		}

		void Capture(bool transferEnable)
		{
			if (mBuffer == NULL) {
				mBuffer = new TS::Buffer(mDevice);
			}

			if (transferEnable) {
				mBuffer->Run();
			} else {
				mBuffer->Stop();
			}
		}

		void Menu2()
		{
			::wprintf(L"[foCXj[ (2^3y[W)]\n");
			::wprintf(L"0: PCI NbNJE^擾\n");
			::wprintf(L"1: PCI CeV^C}擾\n");
			::wprintf(L"2: PCI CeV^C}ݒ\n");
			::wprintf(L"3: `[i[ȓd͐ݒ\n");
			::wprintf(L"4: Xg[Lݒ\n");
			::wprintf(L"5: `lݒ~4\n");
			::wprintf(L"6:  1y[W\n");
			::wprintf(L">");
			uint number = GetNumber(6);
			switch (number) {
				case 0: GetPciClockCounter(); break;
				case 1: GetPciLatencyTimer(); break;
				case 2: SetPciLatencyTimer(); break;
				case 3: SetTunerSleep(); break;
				case 4: SetStreamEnable(); break;
				case 5: ScanTest(); break;
			}
		}

		void GetPciClockCounter()
		{
			uint counter;
			status status = mDevice->GetPciClockCounter(&counter);
			if (Error(L"Device::GetPciClockCounter()", status)) return;

			::wprintf(L"PCI NbNJE^: 0x%08x\n", counter);
		}

		void GetPciLatencyTimer()
		{
			byte latencyTimer;
			status status = mDevice->GetPciLatencyTimer(&latencyTimer);
			if (Error(L"Device::GetPciLatencyTimer()", status)) return;

			::wprintf(L"PCI CeVJE^: 0x%02x\n", latencyTimer);
		}

		void SetPciLatencyTimer()
		{
			::wprintf(L"ݒl͂ĂB(͈:0x00`0xff)\n");
			::wprintf(L">0x");
	
			uint latencyTimer;
			::wscanf_s(L"%x", &latencyTimer);

			if (0xff < latencyTimer) {
				latencyTimer = 0xff;
			}

			status status = mDevice->SetPciLatencyTimer(static_cast<byte>(latencyTimer));
			if (Error(L"Device::SetPciLatencyTimer()", status)) return;
		}

		void SetTunerSleep()
		{
			::wprintf(L"0: `[i[ԍ0\n");
			::wprintf(L"1: `[i[ԍ1\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int tuner = GetNumber(2);
			if (tuner == 2) return;

			::wprintf(L"0: ISDB-S\n");
			::wprintf(L"1: ISDB-T\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int isdb = GetNumber(2);
			if (isdb == 2) return;

			bool boolSleep;
			status status = mDevice->GetTunerSleep(tuner, static_cast<PT::Device::ISDB>(isdb), &boolSleep);
			if (Error(L"Device::GetTunerSleep()", status)) return;

			const wchar_t *LABEL[] = { L"", L"L" };

			for (int i=0; i<sizeof(LABEL)/sizeof(*LABEL); i++) {
				::wprintf(L"%d: %s%s\n", i, LABEL[i], (i==static_cast<int>(boolSleep)) ? L" (ݒl)" : L"");
			}
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int intSleep = GetNumber(2);
			if (intSleep == 2) return;

			boolSleep = intSleep ? true : false;
			status = mDevice->SetTunerSleep(tuner, static_cast<PT::Device::ISDB>(isdb), boolSleep);
			if (Error(L"Device::SetTunerSleep()", status)) return;
		}

		void SetStreamEnable()
		{
			::wprintf(L"0: `[i[ԍ0\n");
			::wprintf(L"1: `[i[ԍ1\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int tuner = GetNumber(2);
			if (tuner == 2) return;

			::wprintf(L"0: ISDB-S\n");
			::wprintf(L"1: ISDB-T\n");
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int isdb = GetNumber(2);
			if (isdb == 2) return;

			bool boolEnable;
			status status = mDevice->GetStreamEnable(tuner, static_cast<PT::Device::ISDB>(isdb), &boolEnable);
			if (Error(L"Device::GetStreamEnable()", status)) return;

			const wchar_t *LABEL[] = { L"", L"L" };

			for (int i=0; i<sizeof(LABEL)/sizeof(*LABEL); i++) {
				::wprintf(L"%d: %s%s\n", i, LABEL[i], (i==static_cast<int>(boolEnable)) ? L" (ݒl)" : L"");
			}
			::wprintf(L"2: LZ\n");
			::wprintf(L">");
			int intEnable = GetNumber(2);
			if (intEnable == 2) return;

			boolEnable = intEnable ? true : false;
			status = mDevice->SetStreamEnable(tuner, static_cast<PT::Device::ISDB>(isdb), boolEnable);
			if (Error(L"Device::SetStreamEnable()", status)) return;
		}

		status ScanS(uint tuner, uint startChannel, uint endChannel)
		{
			status status = PT::STATUS_OK;

			for (uint i=startChannel; i<=endChannel; i++) {
				bool bs;
				uint number;
				GetChannelNameS(i, &bs, &number);
				::wprintf(L"(S%3d) `l:%s%02d", i, bs ? L"BS" : L"ND", number);
		  
				status = mDevice->SetFrequency(tuner, PT::Device::ISDB_S, i);
				if (Error(L"Device::SetFrequency()", status)) return status;

				PT::Device::TmccS tmcc;
				::memset(&tmcc, 0, sizeof(tmcc));
				for (uint j=0; j<6; j++) {
					status = mDevice->GetTmccS(tuner, &tmcc);
					if (status == PT::STATUS_OK) break;
					OS::Thread::Sleep((50)*1000*1000);
				}
				bool tmccIsValid = (status == PT::STATUS_OK);

				uint cn, currentAgc, maxAgc;
				status = mDevice->GetCnAgc(tuner, PT::Device::ISDB_S, &cn, &currentAgc, &maxAgc);
				if (Error(L"Device::GetCnAgc()", status)) return status;

				::wprintf(L" CN:%02d.%02ddB AGC:%03d/%03d", cn / 100, cn % 100, currentAgc, maxAgc);

				if (tmccIsValid) {
					::wprintf(L"\n");
				} else {
					::wprintf(L" TMCC Ms\n");
				}
	
				if (tmccIsValid) {
					for (uint j=0; j<8; j++) {
						uint id = tmcc.Id[j];
						if (id != 0xffff) {
							::wprintf(L"       ");
							::wprintf(L"TS-ID %04x", id);
							mDevice->SetIdS(tuner, id);
							for (uint k=0; k<5; k++) {
								uint currentId;
								mDevice->GetIdS(tuner, &currentId);
								if (currentId == id) break;
								OS::Thread::Sleep((10)*1000*1000);
							}
							PT::Device::LayerS layer;
							mDevice->GetLayerS(tuner, &layer);
							for (uint i=0; i<2; i++) {
								::wprintf(L" %s", (i == 0) ? L"Kw" : L"Kw");
								::wprintf(L" [h:%d Xbg:%02d", layer.Mode[i], layer.Count[i]);
							}
							::wprintf(L"\n");
						}
					}

					PrintFrequencyOffset(tuner, PT::Device::ISDB_S);
				}
			}

			return status;
		}

		status ScanT(uint tuner, uint startChannel, uint endChannel)
		{
			status status = PT::STATUS_OK;

			for (uint i=startChannel; i<=endChannel; i++) {
				bool catv;
				uint number;
				GetChannelNameT(i, &catv, &number);
				::wprintf(L"(T%3d) `l:%s%2d", i, catv ? L"C" : L" ", number);

				status = mDevice->SetFrequency(tuner, PT::Device::ISDB_T, i);
				if (Error(L"Device::SetFrequency", status)) return status;

				PT::Device::TmccT tmcc;
				::memset(&tmcc, 0, sizeof(tmcc));

				status = mDevice->GetTmccT(tuner, &tmcc);
				bool tmccIsValid = (status == PT::STATUS_OK);

				uint cn, currentAgc, maxAgc;
				status = mDevice->GetCnAgc(tuner, PT::Device::ISDB_T, &cn, &currentAgc, &maxAgc);
				if (Error(L"Device::GetCnAgc()", status)) return status;

				bool locked[3];
				status = mDevice->GetLockedT(tuner, locked);
				if (Error(L"Device::GetLockedT()", status)) return status;

				::wprintf(L" CN:%02d.%02ddB AGC:%03d/%03d bN:%d%d%d",
					cn / 100, cn % 100,
					currentAgc, maxAgc,
					locked[0], locked[1], locked[2]);

				if (tmccIsValid) {
					::wprintf(L" VXe:%d", tmcc.System);
					::wprintf(L"\n");
					for (uint j=0; j<3; j++) {
						::wprintf(L"      ");
						::wprintf(L" %cKw", L'A'+j);

						::wprintf(L" ϒ:");
						uint mode = tmcc.Mode[j];
						if (mode != 7) {
							::wprintf(L"%d", mode);
						} else {
							::wprintf(L"-");
						}

						::wprintf(L" :");
						uint rate = tmcc.Rate[j];
						if (mode != 7) {
							::wprintf(L"%d", rate);
						} else {
							::wprintf(L"-");
						}

						::wprintf(L" C^[[u:");
						uint interleave = tmcc.Interleave[j];
						if (mode != 7) {
							::wprintf(L"%d", interleave);
						} else {
							::wprintf(L"-");
						}

						::wprintf(L" ZOg:");
						uint segment = tmcc.Segment[j];
						if (segment != 15) {
							::wprintf(L"%02d", segment);
						} else {
							::wprintf(L"--");
						}

						::wprintf(L"\n");
					}

					PrintFrequencyOffset(tuner, PT::Device::ISDB_T);

				} else {
					::wprintf(L" TMCC Ms\n");
				}
			}

			return status;
		}

		void PrintFrequencyOffset(uint tuner, PT::Device::ISDB isdb)
		{
			int offset[2];
			mDevice->GetFrequencyOffset(tuner, isdb, &offset[0], &offset[1]);

			bool minus[2];
			
			for (uint k=0; k<2; k++) {
				minus[k] = (offset[k] < 0);
				if (minus[k]) {
					offset[k] *= -1;
				}
			}

			::wprintf(L"       ");
			::wprintf(L"NbNg덷:%s%d.%02dppm", minus[0] ? L"-" : L"+", offset[0] / 100, offset[0] % 100);
			::wprintf(L" / ");

			wchar_t *si;
			if (isdb == PT::Device::ISDB_S) {
				offset[1] /= 1000;
				si = L"k";
			} else {
				si = L"";
			}
			::wprintf(L"LAg덷:%s%d%sHz" , minus[1] ? L"-" : L"+", offset[1], si);
			::wprintf(L"\n");
		}

		void ScanTest()
		{
			ScanS(0, 1, 1);
			ScanS(1, 7, 7);

			ScanT(0, 72, 72);
			ScanT(1, 73, 73);
		}

		void GetChannelNameS(uint channel, bool *bs, uint *number) const
		{
			_ASSERTE(bs);
			_ASSERTE(number);

			if (channel < 12) {
				*bs = true;
				*number = 1 + 2*channel;
			} else if (channel < 24) {
				channel -= 12;

				*bs = false;
				*number = 2 + 2*channel;
			} else {
				channel -= 24;

				*bs = false;
				*number = 1 + 2*channel;
			}
		}

		void GetChannelNameT(uint channel, bool *catv, uint *number) const
		{
			static const byte TABLE[][3] = {
				{   2,  0,   3 },
				{  12,  1,  22 },
				{  21,  0,  12 },
				{  62,  1,  63 },
				{ 112,  0,  62 }
			};

			for (uint i=0; i<sizeof(TABLE)/sizeof(*TABLE); i++) {
				if (channel <= TABLE[i][0]) {
					_ASSERTE(catv); *catv = TABLE[i][1] ? true : false;
					_ASSERTE(number); *number = channel + TABLE[i][2] - TABLE[i][0];
					break;
				}
			}
		}
	};
}
}

#endif
